<?php
/**
 * Bitrix Framework
 * @package bitrix
 * @subpackage crm
 * @copyright 2001-2016 Bitrix
 */
namespace Bitrix\Crm\WebForm;

use Bitrix\Crm\WebForm\Internals\FieldDependenceTable;
use Bitrix\Crm\WebForm\Internals\FormStartEditTable;
use Bitrix\Crm\WebForm\Internals\FormTable;
use Bitrix\Crm\WebForm\Internals\FieldTable;
use Bitrix\Crm\WebForm\Internals\FormViewTable;
use Bitrix\Crm\WebForm\Internals\FormCounterTable;
use Bitrix\Crm\WebForm\Internals\PresetFieldTable;
use Bitrix\Main\Application;
use Bitrix\Main\EventManager;
use Bitrix\Main\Event;
use Bitrix\Main\Loader;
use Bitrix\Main\Localization\Loc;
use Bitrix\Main\Entity\Result as EntityResult;
use Bitrix\Main\Type\DateTime;
use Bitrix\Main\Mail\Event as MailEvent;
use Bitrix\Main\Context;
use Bitrix\Main\Config\Option;

Loc::loadMessages(__FILE__);

class Form
{
	protected $id = null;
	protected static $defaultParams = array(
		'ACTIVE' => 'N',
		'IS_SYSTEM' => 'N',
		'COPYRIGHT_REMOVED' => 'N',
		'FIELDS' => array(),
		'PRESET_FIELDS' => array(),
		'INVOICE_SETTINGS' => array(),
		'FORM_SETTINGS' => array(),
		'DEPENDENCIES' => array()
	);
	protected $params = array();
	protected $errors = array();

	public function __construct($id = null, array $params = null)
	{
		$this->params = self::$defaultParams;

		if($id)
		{
			$this->load($id);
		}

		if($params)
		{
			$this->set($params);
		}
	}

	public function isSystem()
	{
		return $this->params['IS_SYSTEM'] == 'Y';
	}

	public function setSystem()
	{
		$this->params['IS_SYSTEM'] = 'Y';
	}

	public function set(array $params)
	{
		$this->params = $params;
	}

	public function get()
	{
		return $this->params;
	}

	public function merge($params)
	{
		$this->set($params + $this->get());
	}

	public static function getIdByCode($formCode)
	{
		$idByCode = array();
		$cacheId = 'crm_webform_getIdByCode_' . serialize($formCode);
		$cache = \Bitrix\Main\Data\Cache::createInstance();
		if($cache->startDataCache(36000, $cacheId))
		{
			$formDb = FormTable::getList(array(
				'select' => array('ID', 'CODE'),
				'filter' => array('=CODE' => $formCode),
				'limit' => 1
			));
			while($form = $formDb->fetch())
			{
				$idByCode[$form['CODE']] = $form['ID'];
			}

			if(isset($idByCode[$formCode]))
			{
				if(defined("BX_COMP_MANAGED_CACHE"))
				{
					global $CACHE_MANAGER;
					$CACHE_MANAGER->startTagCache($cache->getPath($cacheId));
					$CACHE_MANAGER->RegisterTag(Form::getCacheTag($idByCode[$formCode]));
				}
			}

			$cache->endDataCache(array('CODE_BY_ID' => $idByCode));

			if(isset($idByCode[$formCode]))
			{
				if(defined("BX_COMP_MANAGED_CACHE"))
				{
					global $CACHE_MANAGER;
					$CACHE_MANAGER->endTagCache();
				}
			}
		}
		else
		{
			$cacheVars = $cache->getVars();
			if(isset($cacheVars['CODE_BY_ID']))
			{
				$idByCode = $cacheVars['CODE_BY_ID'];
			}
		}

		return isset($idByCode[$formCode]) ? $idByCode[$formCode] : null;
	}

	public static function updateBackgroundImage($formId, $fileId)
	{
		$updateResult = FormTable::update($formId, array('BACKGROUND_IMAGE' => $fileId));
		return $updateResult->isSuccess();
	}

	public static function delete($formId, $forceSystem = false)
	{
		$form = Internals\FormTable::getRowById($formId);
		if(!$form || (!$forceSystem && $form['IS_SYSTEM'] == 'Y'))
		{
			return false;
		}

		$deleteResult = Internals\FormTable::delete($formId);
		if($deleteResult->isSuccess())
		{
			static::cleanCacheByTag($formId);
			return true;
		}
		else
		{
			return false;
		}
	}

	public function isActive()
	{
		return $this->params['ACTIVE'] == 'Y';
	}

	public function checkSecurityCode($code)
	{
		return $this->params['SECURITY_CODE'] === $code;
	}

	public function loadOnlyForm($id)
	{
		$this->setId($id);
		$result = Internals\FormTable::getRowById($id);
		if(!$result)
		{
			return false;
		}

		$this->params = $result;
		return true;
	}

	public function load($id)
	{
		$this->setId($id);
		if(count($this->params) == count(self::$defaultParams) && !$this->loadOnlyForm($this->id))
		{
			return false;
		}

		$this->params['PRESET_FIELDS'] = array();
		$dbPresetField = PresetFieldTable::getList(array(
			'select' => array('ENTITY_NAME', 'FIELD_NAME', 'VALUE'),
			'filter' => array('=FORM_ID' => $id)
		));
		while($presetField = $dbPresetField->fetch())
		{
			$this->params['PRESET_FIELDS'][] = $presetField;
		}

		$this->params['FIELDS'] = FieldTable::getList(array(
			'filter' => array('=FORM_ID' => $id),
			'order' => array('SORT' => 'ASC', 'CAPTION')
		))->fetchAll();


		$this->params['DEPENDENCIES'] = Internals\FieldDependenceTable::getList(array(
			'filter' => array('=FORM_ID' => $id)
		))->fetchAll();

		return true;
	}

	public function save($onlyCheck = false)
	{
		$this->errors = array();
		$result = $this->params;

		$scripts = $result['SCRIPTS'];
		unset($result['SCRIPTS']);

		$fields = $result['FIELDS'];
		unset($result['FIELDS']);

		$dependencies = $result['DEPENDENCIES'];
		unset($result['DEPENDENCIES']);

		$presetFields = $result['PRESET_FIELDS'];
		unset($result['PRESET_FIELDS']);

		if($onlyCheck)
		{
			if(!in_array($result['ENTITY_SCHEME'], $this->getAllowedEntitySchemes()))
			{
				$this->errors[] = Loc::getMessage('CRM_WEBFORM_FORM_ERROR_SCHEME');
			}

			$formResult = new EntityResult;
			$result['DATE_CREATE'] = new DateTime();
			Internals\FormTable::checkFields($formResult, $this->id, $result);
			$this->prepareResult('FIELDS', $formResult);

			foreach($presetFields as $presetField)
			{
				$presetField['FORM_ID'] = (int) $this->id;
				$presetFieldResult = new EntityResult;
				Internals\PresetFieldTable::checkFields($presetFieldResult, null, $presetField);
				$replaceList = null;
				if(!$presetFieldResult->isSuccess())
				{
					$field = EntityFieldProvider::getField($presetField['ENTITY_NAME'] . '_' . $presetField['FIELD_NAME']);
					if($field)
					{
						$replaceList = array('VALUE' => $field['caption']);
					}
				}
				$this->prepareResult('PRESET_FIELDS', $presetFieldResult, $replaceList);
			}

			$fieldCodeList = array();
			foreach($fields as $field)
			{
				$field['FORM_ID'] = (int) $this->id;

				$fieldResult = new EntityResult;
				Internals\FieldTable::checkFields($fieldResult, null, $field);
				$this->prepareResult('FIELDS', $fieldResult);

				$fieldCodeList[] = $field['CODE'];
			}

			foreach($dependencies as $dependency)
			{
				$dependency['FORM_ID'] = (int) $this->id;

				if(!in_array($dependency['IF_FIELD_CODE'], $fieldCodeList))
				{
					continue;
				}
				if(!in_array($dependency['DO_FIELD_CODE'], $fieldCodeList))
				{
					continue;
				}

				$dependencyResult = new EntityResult;
				Internals\FieldDependenceTable::checkFields($dependencyResult, null, $dependency);
				$this->prepareResult('DEPENDENCIES', $dependencyResult);
			}

			return;
		}

		if(!$this->check())
		{
			return;
		}

		if($this->id)
		{
			$formResult = Internals\FormTable::update($this->id, $result);
		}
		else
		{
			$result['DATE_CREATE'] = new DateTime();
			$formResult = Internals\FormTable::add($result);
			$this->id = $formResult->getId();
		}

		if(!$formResult->isSuccess())
		{
			return;
		}

		/* PRESET FIELDS */
		Internals\PresetFieldTable::delete(array('FORM_ID' => $this->id));
		foreach($presetFields as $presetField)
		{
			$presetFieldResult = Internals\PresetFieldTable::add(array(
				'ENTITY_NAME' => $presetField['ENTITY_NAME'],
				'FIELD_NAME' => $presetField['FIELD_NAME'],
				'VALUE' => $presetField['VALUE'],
				'FORM_ID' => $this->id
			));
			$this->prepareResult('PRESET_FIELDS', $presetFieldResult);
		}


		/* FIELDS */
		$existedFieldList = array();
		$existedFieldDb = Internals\FieldTable::getList(array(
			'select' => array('ID'),
			'filter' => array('=FORM_ID' => $this->id)
		));
		while($existedField = $existedFieldDb->fetch())
		{
			$existedFieldList[] = $existedField['ID'];
		}

		$newFieldList = array();
		foreach($fields as $field)
		{
			$field['FORM_ID'] = $this->id;

			if($field['ID'] > 0)
			{
				$fieldResult = Internals\FieldTable::update($field['ID'], $field);
				$newFieldList[] = $field['ID'];
			}
			else
			{
				$fieldResult = Internals\FieldTable::add($field);
			}

			$this->prepareResult('FIELDS', $fieldResult);
		}

		$deleteFieldList = array_diff($existedFieldList, $newFieldList);
		foreach($deleteFieldList as $deleteFieldId)
		{
			Internals\FieldTable::delete($deleteFieldId);
		}

		/* DEPENDENCIES */
		$fieldCodeList = array();
		$fieldCodeDb = Internals\FieldTable::getList(array(
			'select' => array('CODE'),
			'filter' => array('=FORM_ID' => $this->id)
		));
		while($fieldCode = $fieldCodeDb->fetch())
		{
			$fieldCodeList[] = $fieldCode['CODE'];
		}

		$fieldDepDb = Internals\FieldDependenceTable::getList(array('filter' => array('=FORM_ID' => $this->id)));
		while($fieldDep = $fieldDepDb->fetch())
		{
			Internals\FieldDependenceTable::delete($fieldDep['ID']);
		}
		foreach($dependencies as $dependency)
		{
			if(!in_array($dependency['IF_FIELD_CODE'], $fieldCodeList))
			{
				continue;
			}
			if(!in_array($dependency['DO_FIELD_CODE'], $fieldCodeList))
			{
				continue;
			}

			$dependency['FORM_ID'] = $this->id;

			$dependencyResult = Internals\FieldDependenceTable::add($dependency);
			$this->prepareResult('DEPENDENCIES', $dependencyResult);
		}

		$this->cleanCacheByTag($this->id);
	}

	public function getErrors()
	{
		return $this->errors;
	}

	public function hasErrors()
	{
		return count($this->errors) > 0;
	}

	protected function prepareResult($sect, EntityResult $entityResult, $replaceList = null)
	{
		if($entityResult->isSuccess())
		{
			return;
		}

		$errors = $entityResult->getErrors();
		foreach($errors as $error)
		{
			$errorMessage = $error->getMessage();
			if($replaceList)
			{
				$errorMessage = str_replace(array_keys($replaceList), array_values($replaceList), $errorMessage);
			}
			$this->errors[] = $errorMessage;
		}
	}

	public function check()
	{
		$this->save(true);

		return count($this->errors) === 0;
	}

	public function setId($id)
	{
		$this->id = $id;
	}

	public function getId()
	{
		return $this->id;
	}

	public function getButtonCaption()
	{
		return $this->params['BUTTON_CAPTION'] ? $this->params['BUTTON_CAPTION'] : Loc::getMessage('CRM_WEBFORM_FORM_BUTTON_CAPTION_DEFAULT');
	}

	public function getFields()
	{
		return $this->params['FIELDS'];
	}

	public function getAllowedEntitySchemes()
	{
		//TODO: fields checker
		return array_keys(Entity::getSchemes());
		//return array(Helper::ENUM_ENTITY_SCHEME_CONTACT);
	}

	public function getPresetFields()
	{
		return $this->params['PRESET_FIELDS'];
	}

	public function getDependencies()
	{
		$dependencyList = array();
		foreach($this->params['DEPENDENCIES'] as $dependency)
		{
			$dependencyList[$dependency['DO_FIELD_CODE']][] = array(
				'if' => array(
					'fieldname' => $dependency['IF_FIELD_CODE'],
					'action' => $dependency['IF_ACTION'],
					'value' => $dependency['IF_VALUE'],
				),
				'do' => array(
					'action' => $dependency['DO_ACTION'],
					'value' => $dependency['DO_VALUE'],
				),
			);

			// add mirror dependency
			if($dependency['IF_ACTION'] != 'change')
			{
				continue;
			}

			if(!in_array($dependency['DO_ACTION'], array('show', 'hide')))
			{
				continue;
			}

			$mirror = $dependency['DO_ACTION'] == 'hide' ? 'show' : 'hide';
			$dependencyList[$dependency['DO_FIELD_CODE']][] = array(
				'if' => array(
					'fieldname' => $dependency['IF_FIELD_CODE'],
					'action' => $dependency['IF_ACTION'],
					'value' => $dependency['IF_VALUE'],
					'operation' => '!='
				),
				'do' => array(
					'action' => $mirror,
				),
			);
		}

		return $dependencyList;
	}

	public function getFieldsDescription()
	{
		return EntityFieldProvider::getFieldsDescription($this->getFields());
	}

	public function getFieldsMap()
	{
		$fields = $this->getFieldsDescription();
		$dependencyList = $this->getDependencies();
		$currencyId = $this->getCurrencyId();

		$fieldList = array();
		foreach($fields as $field)
		{

			if($field['TYPE'] == 'section')
			{
				$preparedField = array(
					'type' => $field['TYPE'],
					'name' => $field['CODE'],
					'caption' => $field['CAPTION'],
				);
			}
			else
			{
				$preparedField = array(
					'type' => $field['TYPE'],
					'type_original' => $field['TYPE_ORIGINAL'],
					'name' => $field['CODE'], // 'wf_field_' . $field['ID'],
					'entity_name' => $field['ENTITY_NAME'],
					'entity_field_name' => $field['ENTITY_FIELD_NAME'],
					'caption' => $field['CAPTION'] ? $field['CAPTION'] : $field['ENTITY_FIELD_CAPTION'],
					'required' => $field['REQUIRED'] == 'Y' ? true : false,
					'multiple' => $field['MULTIPLE'] == 'Y' ? true : false,
					'multiple_original' => $field['MULTIPLE_ORIGINAL'],
					'hidden' => false,
					'placeholder' => $field['PLACEHOLDER'],
				);

				if(isset($field['ITEMS']))
				{
					$preparedField['items'] = array();
					foreach($field['ITEMS'] as $item)
					{
						$preparedField['items'][] = array(
							'title' => $item['VALUE'],
							'value' => $item['ID'],
							'price' => $item['PRICE'] ? $item['PRICE'] : null,
							'price_formatted' => $item['PRICE'] ?
								\CCrmCurrency::MoneyToString($item['PRICE'], $currencyId)
								:
								null
						);
					}
				}
			}

			if(isset($dependencyList[$field['CODE']]))
			{
				$preparedField['dependences'] = $dependencyList[$field['CODE']];
				$preparedField['hidden'] = true;
			}

			$fieldList[] = $preparedField;
		}

		return $fieldList;
	}

	public function getExternalAnalyticsData()
	{
		$data = Helper::getExternalAnalyticsData($this->params['CAPTION']);
		$steps = array();

		$steps[] = array(
			'NAME' => $data['view']['name'],
			'CODE' => $data['view']['code']
		);
		$steps[] = array(
			'NAME' => $data['start']['name'],
			'CODE' => $data['start']['code']
		);
		foreach($this->getFieldsMap() as $field)
		{
			if(in_array($field['type'], array('hr', 'br', 'section')))
			{
				continue;
			}

			$steps[] = array(
				'NAME' => str_replace('%name%', $field['caption'], $data['field']['name']),
				'CODE' => str_replace('%code%', $field['name'], $data['field']['code']),
			);
		}
		$steps[] = array(
			'NAME' => $data['end']['name'],
			'CODE' => $data['end']['code']
		);

		foreach($steps as $stepIndex => $step)
		{
			$step['NAME'] = str_replace('%name%', $step['NAME'], $data['template']['name']);
			$step['CODE'] = str_replace('%code%', $step['CODE'], $data['template']['code']);
			$steps[$stepIndex] = $step;
		}

		return array(
			'CATEGORY' => $data['category'],
			'STEPS' => $steps
		);
	}

	public function getCurrencyId()
	{
		return $this->params['CURRENCY_ID'] ? $this->params['CURRENCY_ID'] : \CCrmCurrency::GetBaseCurrencyID();
	}

	public static function copy($formId, $userId = null)
	{
		// copy form
		$form = FormTable::getRowById($formId);
		if(!$form)
		{
			return null;
		}

		unset($form['ID'], $form['DATE_CREATE'], $form['ACTIVE_CHANGE_DATE'], $form['SECURITY_CODE'], $form['XML_ID']);
		$form['NAME'] = Loc::getMessage('CRM_WEBFORM_FORM_COPY_NAME_PREFIX') . ' ' . $form['NAME'];
		$form['ACTIVE'] = 'N';
		$form['IS_SYSTEM'] = 'N';
		$form['ACTIVE_CHANGE_BY'] = $userId;
		$form['DATE_CREATE'] = new DateTime();
		$resultFormAdd = FormTable::add($form);
		if(!$resultFormAdd->isSuccess())
		{
			return null;
		}
		$newFormId = $resultFormAdd->getId();

		// copy fields
		$fieldDb = FieldTable::getList(array(
			'filter' => array('=FORM_ID' => $formId)
		));
		while($field = $fieldDb->fetch())
		{
			unset($field['ID']);
			$field['FORM_ID'] = $newFormId;
			FieldTable::add($field);
		}

		// copy field dependencies
		$fieldDepDb = FieldDependenceTable::getList(array(
			'filter' => array('=FORM_ID' => $formId)
		));
		while($fieldDep = $fieldDepDb->fetch())
		{
			unset($fieldDep['ID']);
			$fieldDep['FORM_ID'] = $newFormId;
			FieldDependenceTable::add($fieldDep);
		}

		// copy preset fields
		$presetFieldDb = PresetFieldTable::getList(array(
			'filter' => array('=FORM_ID' => $formId)
		));
		while($presetField = $presetFieldDb->fetch())
		{
			$presetField['FORM_ID'] = $newFormId;
			PresetFieldTable::add($presetField);
		}


		return $newFormId;
	}

	public static function activate($formId, $isActivate = true, $changeUserBy = null)
	{
		$updateFields = array('ACTIVE' => $isActivate ? 'Y' : 'N');
		if($changeUserBy)
		{
			$updateFields['ACTIVE_CHANGE_BY'] = $changeUserBy;
		}
		$updateResult = FormTable::update($formId, $updateFields);
		if($updateResult->isSuccess())
		{
			static::cleanCacheByTag($formId);
			return true;
		}
		else
		{
			return false;
		}
	}

	/*
	 * @return Result
	 * */
	public function addResult($resultFields)
	{
		$this->errors = array();

		// get section fields
		$sectionFields = array();
		$currentSection = '';
		$fieldsMap = $this->getFieldsMap();
		foreach($fieldsMap as $field)
		{
			if($field['type'] == 'section')
			{
				$currentSection = $field['name'];
				continue;
			}

			if(!$currentSection)
			{
				continue;
			}

			$sectionFields[$currentSection][] = $field['name'];
		}

		// format fields by name
		$fields = array();
		foreach($resultFields as $fieldKey => $field)
		{
			$fields[$field['name']] = $field;
		}

		$hiddenFieldNames = array();
		// set hidden flag
		foreach($resultFields as $fieldKey => $field)
		{
			if(!isset($field['dependences']))
			{
				$field['hidden'] = false;
				$data['FIELDS'][$fieldKey] = $field;
				continue;
			}

			$isHidden = false;
			foreach($field['dependences'] as $dep)
			{
				if($dep['if']['action'] != 'change')
				{
					continue;
				}

				if(!isset($fields[$dep['if']['fieldname']]))
				{
					continue;
				}

				$valuesA = $fields[$dep['if']['fieldname']]['values'];
				$valueB = $dep['if']['value'];
				$isSuccess = false;
				switch($dep['if']['operation'])
				{
					case '!=':
						$isSuccess = !in_array($valueB, $valuesA);
						break;

					default:
						$isSuccess = in_array($valueB, $valuesA);
						break;
				}

				if(!$isSuccess)
				{
					continue;
				}

				$isHidden = $dep['do']['action'] == 'hide';
			}

			if($field['type'] == 'section')
			{
				foreach($sectionFields[$field['name']] as $sectionFieldName)
				{
					$hiddenFieldNames[$sectionFieldName][] = $isHidden;
				}
			}
			else
			{
				$field['hidden'] = $isHidden;
				$resultFields[$fieldKey] = $field;
			}
		}

		$fieldEmailValue = null;
		$fieldPhoneValue = null;
		$fieldPhoneEntityTypeName = null;
		foreach($resultFields as $fieldKey => $field)
		{
			if(($field['entity_field_name'] == 'EMAIL' || $field['type'] == 'email') && $field['values'][0])
			{
				$fieldEmailValue = $field['values'][0];
			}
			if(($field['entity_field_name'] == 'PHONE' || $field['type'] == 'phone') && $field['values'][0])
			{
				$fieldPhoneEntityTypeName = $field['entity_name'];
				$fieldPhoneValue = $field['values'][0];
			}

			if(!isset($hiddenFieldNames[$field['name']]))
			{
				continue;
			}

			$field['hidden'] = end($hiddenFieldNames[$field['name']]);
			$resultFields[$fieldKey] = $field;
		}

		$resultProducts = array();
		$activityFields = array();
		foreach($resultFields as $fieldKey => $field)
		{
			if($field['hidden'])
			{
				continue;
			}

			if(!isset($field['values'][0]) || !$field['values'][0])
			{
				continue;
			}

			$activityFieldValues = array();
			if(is_array($field['items']) && count($field['items']) > 0)
			{
				foreach($field['items'] as $item)
				{
					//TODO: CHECK [0]
					if($item['value'] != $field['values'][0])
					{
						continue;
					}

					$activityFieldValues[] = $item;
				}
			}
			else
			{
				$activityFieldValues = $field['values'];
			}
			$activityFields[] = array(
				'type' => $field['type'],
				'required' => $field['required'],
				'caption' => $field['caption'],
				'value' => $activityFieldValues,
			);


			if($field['type'] != 'product')
			{
				continue;
			}

			foreach($field['items'] as $item)
			{
				//TODO: CHECK [0]
				if($item['value'] != $field['values'][0])
				{
					continue;
				}

				$resultProducts[] = array(
					'ID' => is_numeric($item['value']) ? $item['value'] : 0,
					'NAME' => $item['title'],
					'PRICE' => $item['price'],
				);
			}
		}


		// add Result
		$data = array(
			'FORM' => $this->params,
			'DUPLICATE_MODE' => $this->params['DUPLICATE_MODE'],
			'PRESET_FIELDS' => $this->params['PRESET_FIELDS'],
			'ENTITY_SCHEME' => $this->params['ENTITY_SCHEME'],
			'INVOICE_SETTINGS' => $this->params['INVOICE_SETTINGS'],
			'FORM_ID' => $this->id,
			'FIELDS' => $resultFields,
			'ASSIGNED_BY_ID' => $this->params['ASSIGNED_BY_ID'],
			'PRODUCTS' => $resultProducts,
			'ACTIVITY_FIELDS' => $activityFields,
		);
		$result = new Result(null, $data);
		$result->save();
		if($result->hasErrors())
		{
			$this->errors = $result->getErrors();
		}
		else
		{
			if($fieldEmailValue)
			{
				self::sendEventFormSent(array(
						'RESULT_SUCCESS_TEXT' => $this->params['RESULT_SUCCESS_TEXT'],
						'RESULT_SUCCESS_URL' => $this->params['RESULT_SUCCESS_URL'],
						'RESULT_FAILURE_TEXT' => $this->params['RESULT_FAILURE_TEXT'],
						'RESULT_FAILURE_URL' => $this->params['RESULT_FAILURE_URL'],
						'FORM_CAPTION' => $this->params['CAPTION'],
						'EMAIL_TO' => $fieldEmailValue,
					)
				);
			}
			if($fieldPhoneValue && $this->params['IS_CALLBACK_FORM'] == 'Y')
			{
				if(Loader::includeModule('voximplant') && count(\CVoxImplantConfig::GetPortalNumbers(true)) > 0)
				{
					$callEvent = new Event(
						'crm',
						'OnCrmCallbackFormSubmitted',
						array(
							array(
								'CRM_ENTITY_TYPE' => $fieldPhoneEntityTypeName,
								'CRM_ENTITY_ID' => $result->getResultEntity()->getEntityIdByTypeName($fieldPhoneEntityTypeName),
								'CALL_FROM' => $this->params['CALL_FROM'],
								'CALL_TO' => $fieldPhoneValue,
								'TEXT' => $this->params['CALL_TEXT'],
							)
						)
					);
					EventManager::getInstance()->send($callEvent);
				}
			}
		}

		return $result;
	}

	protected function sendEventFormSent($fields)
	{
		MailEvent::send(array(
				'EVENT_NAME' => 'CRM_WEB_FORM_FILLED',
				'C_FIELDS' => $fields,
				'LID' => Context::getCurrent()->getSite()
		));
	}

	public function getScripts($publicFormPath)
	{
		$script = new Script(
			Context::getCurrent()->getServer()->getHttpHost(),
			Context::getCurrent()->getRequest()->isHttps()
		);

		$scriptParams = array(
			'id' => $this->params['ID'],
			'lang' => Context::getCurrent()->getLanguage(),
			'sec' => $this->params['SECURITY_CODE']
		);

		return array(
			'INLINE' => $script->getInline($scriptParams),
			'BUTTON' => $script->getButton($scriptParams + array('button_caption' => Loc::getMessage('CRM_WEBFORM_FORM_SCRIPT_BUTTON_TEXT'))),
			'LINK' => $script->getLink($scriptParams + array('button_caption' => Loc::getMessage('CRM_WEBFORM_FORM_SCRIPT_BUTTON_TEXT'))),
			'DELAY' => $script->getDelay($scriptParams + array('delay' => 5))
		);
	}

	public function sendScriptsEmail($email)
	{

		$this->arResult = array(
			'FORM_NAME' => $formData['CAPTION'] ? $formData['CAPTION'] : $formData['NAME'],
			'LINK' => str_replace('#form_id#', $formData['ID'], $linkPath),
			'SCRIPTS' => array(
				'INLINE' => $script->getInline($scriptParams),
				'BUTTON' => $script->getButton($scriptParams + array('button_caption' => Loc::getMessage('CRM_WEBFORM_FORM_SCRIPT_BUTTON_TEXT'))),
				'LINK' => $script->getLink($scriptParams + array('button_caption' => Loc::getMessage('CRM_WEBFORM_FORM_SCRIPT_BUTTON_TEXT'))),
				'DELAY' => $script->getDelay($scriptParams + array('delay' => 5))
			),
		);
		$fields = array(
			'SCRIPT_DELAY' => $this->params['RESULT_SUCCESS_TEXT'],
			'SCRIPT_INLINE' => $this->params['RESULT_SUCCESS_URL'],
			'SCRIPT_' => $this->params['RESULT_FAILURE_TEXT'],
			'RESULT_FAILURE_URL' => $this->params['RESULT_FAILURE_URL'],
			'FORM_CAPTION' => $this->params['CAPTION'],
			'FORM_NAME' => $this->params['NAME'],
			'EMAIL_TO' => $email,
		);
		MailEvent::send(array(
				'EVENT_NAME' => 'CRM_WEB_FORM_FILLED',
				'C_FIELDS' => $fields,
				'LID' => Context::getCurrent()->getSite()
		));
	}

	public static function getCacheTag($formId)
	{
		return 'BX_CRM_WEBFORM_ID_' . $formId;
	}

	public static function cleanCacheByTag($formId)
	{
		if(defined("BX_COMP_MANAGED_CACHE"))
		{
			$taggedCache = Application::getInstance()->getTaggedCache();
			$taggedCache->clearByTag(static::getCacheTag($formId));
		}
	}

	public static function getCounters($formId, $formEntityScheme = null)
	{
		$result = array(
			'ENTITY' => array(),
			'COMMON' => array()
		);
		$entityList = Entity::getList();
		if($formEntityScheme)
		{
			$entitySchemes = Entity::getSchemes();
			if($entitySchemes[$formEntityScheme])
			{
				foreach($entityList as $entityName => $entityCaption)
				{
					if(!in_array($entityName, $entitySchemes[$formEntityScheme]['ENTITIES']))
					{
						unset($entityList[$entityName]);
					}
				}
			}
		}

		$entityFieldMap = FormCounterTable::getEntityFieldsMap();
		$counters = FormCounterTable::getByFormId($formId);
		foreach($counters as $counter => $value)
		{
			if(isset($entityFieldMap[$counter]))
			{
				$entityName = $entityFieldMap[$counter];
				if(!isset($entityList[$entityName]))
				{
					continue;
				}

				$entityCaption = $entityList[$entityName];
				$result['ENTITY'][] = array(
					'ENTITY_NAME' => $entityName,
					'ENTITY_CAPTION' => $entityCaption,
					'VALUE' => $value,
				);
			}
			else
			{
				$result['COMMON'][$counter] = $value;
			}
		}

		return $result;
	}

	public static function incCounterView($formId)
	{
		FormViewTable::add(array('FORM_ID' => $formId));
		return FormCounterTable::incCounters($formId, array('VIEWS'));
	}

	public static function incCounterStartFill($formId)
	{
		FormStartEditTable::add(array('FORM_ID' => $formId));
		return FormCounterTable::incCounters($formId, array('START_FILL'));
	}

	public static function incCounterEndFill($formId)
	{
		return FormCounterTable::incCounters($formId, array('END_FILL'));
	}

	public static function resetCounters($formId)
	{
		$newCounterId = FormCounterTable::addByFormId($formId);
		// TODO: merge all counters
		return $newCounterId;
	}

	public static function canRemoveCopyright()
	{
		if(!Loader::includeModule('bitrix24'))
		{
			return true;
		}

		if(\CBitrix24::IsDemoLicense())
		{
			return true;
		}

		return \CBitrix24::IsLicensePaid();
	}

	protected static function getMaxActivatedFormLimit()
	{
		return intval(Option::get('crm', '~crm_webform_max_activated', 99999));
	}

	public static function canActivateForm()
	{
		if(!Loader::includeModule("bitrix24"))
		{
			return true;
		}

		$maxActivated = self::getMaxActivatedFormLimit();
		return $maxActivated > FormTable::getCount(array('=ACTIVE' => 'Y', '=IS_SYSTEM' => 'N'));
	}

	public static function actualizeFormsActiveState($maxActivated = null)
	{
		if(!$maxActivated)
		{
			$maxActivated = self::getMaxActivatedFormLimit();
		}

		$formDb = FormTable::getList(array(
			'select' => array('ID'),
			'filter' => array('=ACTIVE' => 'Y', '=IS_SYSTEM' => 'N'),
			'order' => array('ID' => 'ASC')
		));
		while($form = $formDb->fetch())
		{
			if($maxActivated > 0)
			{
				--$maxActivated;
				continue;
			}

			static::activate($form['ID'], false);
		}

		if(!self::canRemoveCopyright())
		{
			$connection = Application::getConnection();
			$connection->query("UPDATE b_crm_webform SET COPYRIGHT_REMOVED='N'");
		}
	}

	public static function onAfterSetOptionCrmWebFormMaxActivated(\Bitrix\Main\Event $event)
	{
		self::actualizeFormsActiveState();
	}

	public static function onBitrix24LicenseChange(\Bitrix\Main\Event $event)
	{
		preg_match("/(project|tf|team)$/is", $event->getParameter(0), $matches);
		$licenseType = strtolower($matches[0]);
		if ($licenseType)
		{
			$maxActivated = null;
			switch($licenseType)
			{
				case 'project':
					$maxActivated = 1;
					break;
				case 'tf':
					$maxActivated = 2;
					break;
				case 'team':
					$maxActivated = 4;
					break;
				case 'demo':
					$maxActivated = 9999;
					break;
			}

			self::actualizeFormsActiveState($maxActivated);
		}
	}
}
