<?php
namespace Bitrix\Rest;

use Bitrix\Main;

/**
 * Class EventTable
 *
 * Fields:
 * <ul>
 * <li> ID int mandatory
 * <li> APP_ID int mandatory
 * <li> EVENT_NAME string(255) mandatory
 * <li> EVENT_HANDLER string(255) mandatory
 * <li> USER_ID int optional
 * </ul>
 *
 * @package Bitrix\Rest
 **/
class EventTable extends Main\Entity\DataManager
{
	const ERROR_UNSUPPORTED_PROTOCOL = 'ERROR_UNSUPPORTED_PROTOCOL';
	const ERROR_WRONG_HANDLER_URL = 'ERROR_WRONG_HANDLER_URL';
	const ERROR_HANDLER_URL_MATCH = 'ERROR_HANDLER_URL_MATCH';
	const ERROR_EVENT_NOT_FOUND = 'ERROR_EVENT_NOT_FOUND';

	/**
	 * Returns DB table name for entity.
	 *
	 * @return string
	 */
	public static function getTableName()
	{
		return 'b_rest_event';
	}

	/**
	 * Returns entity map definition.
	 *
	 * @return array
	 */
	public static function getMap()
	{
		return array(
			'ID' => array(
				'data_type' => 'integer',
				'primary' => true,
				'autocomplete' => true,
			),
			'APP_ID' => array(
				'data_type' => 'integer',
				'required' => true,
			),
			'EVENT_NAME' => array(
				'data_type' => 'string',
				'required' => true,
			),
			'EVENT_HANDLER' => array(
				'data_type' => 'string',
				'required' => true,
			),
			'USER_ID' => array(
				'data_type' => 'integer',
			),
			/**
			 * @deprecated
			 * Use REST_APP
			 */
			'APP' => array(
				'data_type' => 'Bitrix\Bitrix24\AppsTable',
				'reference' => array('=this.APP_ID' => 'ref.ID'),
			),
			'REST_APP' => array(
				'data_type' => 'Bitrix\Rest\AppTable',
				'reference' => array('=this.APP_ID' => 'ref.ID'),
			),
		);
	}

	/**
	 * Removes all application event handlers.
	 *
	 * @param int $appId Application ID.
	 *
	 * @return Main\DB\Result
	 */
	public static function deleteByApp($appId)
	{
		$connection = Main\Application::getConnection();
		return $connection->query("DELETE FROM ".static::getTableName()." WHERE APP_ID='".intval($appId)."'");
	}

	/**
	 * Removes application install event handler.
	 *
	 * @param int $appId Application ID.
	 *
	 * @return Main\DB\Result
	 */
	public static function deleteAppInstaller($appId)
	{
		$connection = Main\Application::getConnection();
		return $connection->query("DELETE FROM ".static::getTableName()." WHERE APP_ID='".intval($appId)."' AND EVENT_NAME='ONAPPINSTALL'");
	}

	/**
	 * Checks event callback URL validity.
	 *
	 * @param string $eventCallback Callback URL.
	 * @param array $appInfo Application info.
	 * @param bool|true $checkInstallUrl Flag, whether to check URL_INSTALL field.
	 *
	 * @return bool
	 *
	 * @throws RestException
	 */
	public static function checkCallback($eventCallback, $appInfo, $checkInstallUrl = true)
	{
		$callbackData = parse_url($eventCallback);

		if(is_array($callbackData)
			&& strlen($callbackData['host']) > 0
			&& strpos($callbackData['host'], '.') > 0
		)
		{
			if($callbackData['scheme'] == 'http' || $callbackData['scheme'] == 'https')
			{
				$host = $callbackData['host'];

				if(strlen($appInfo['URL']) > 0)
				{
					$urlsList = array($appInfo['URL']);

					if(strlen($appInfo['URL_DEMO']) > 0)
					{
						$urlsList[] = $appInfo['URL_DEMO'];
					}

					if($checkInstallUrl && strlen($appInfo['URL_INSTALL']) > 0)
					{
						$urlsList[] = $appInfo['URL_INSTALL'];
					}

					foreach($urlsList as $url)
					{
						$a = parse_url($url);
						if($host == $a['host'] || $a['host'] == 'localhost')
						{
							return true;
						}
					}

					throw new RestException('Handler URL host doesn\'t match application url', static::ERROR_HANDLER_URL_MATCH);
				}
				else
				{
					return true;
				}
			}
			else
			{
				throw new RestException('Unsupported event handler protocol', static::ERROR_UNSUPPORTED_PROTOCOL);
			}
		}
		else
		{
			throw new RestException('Wrong handler URL', static::ERROR_WRONG_HANDLER_URL);
		}
	}

	public static function onBeforeUpdate(Main\Entity\Event $event)
	{
		return static::checkUniq($event);
	}

	public static function onBeforeAdd(Main\Entity\Event $event)
	{
		return static::checkUniq($event);
	}

	protected static function checkUniq(Main\Entity\Event $event)
	{
		$result = new Main\Entity\EventResult();
		$data = $event->getParameter("fields");

		$dbRes = static::getList(array(
			'filter' => array(
				'=APP_ID' => $data['APP_ID'],
				'=EVENT_NAME' => $data['EVENT_NAME'],
				'=EVENT_HANDLER' => $data['EVENT_HANDLER'],
				'=USER_ID' => $data['USER_ID'],
			),
			'select' => array('ID')
		));

		if($dbRes->fetch())
		{
			$result->addError(new Main\Entity\EntityError(
				"Handler already binded"
			));
		}

		return $result;
	}
}
