<?php
/**
 * Bitrix Framework
 * @package bitrix
 * @subpackage rest
 * @copyright 2001-2016 Bitrix
 */

namespace Bitrix\Rest\OAuth;


use Bitrix\Main\Application;
use Bitrix\Rest\OAuthService;

class Auth
{
	const CACHE_TTL = 3600;
	const CACHE_PREFIX = "oauth_";

	const PARAM_LOCAL_USER = 'LOCAL_USER';
	const PARAM_TZ_OFFSET = 'TZ_OFFSET';

	protected static $authQueryParams = array(
		'auth', 'access_token'
	);

	public static function authorizeApplication($clientId, $userId, $state = '')
	{
		if($userId > 0)
		{
			$client = OAuthService::getEngine()->getClient();
			$additionalParams = static::getTokenParams(array(), $userId);

			$codeInfo = $client->getCode($clientId, $state, $additionalParams);

			if($codeInfo['result'])
			{
				return $codeInfo['result'];
			}
			else
			{
				return $codeInfo;
			}
		}

		return false;
	}

	public static function get($appId, $scope, $additionalParams, $userId)
	{
		if($userId > 0)
		{
			$client = OAuthService::getEngine()->getClient();

			$additionalParams = static::getTokenParams($additionalParams, $userId);

			$authResult = $client->getAuth($appId, $scope, $additionalParams);

			if($authResult['result'])
			{
				if($authResult['result']['access_token'])
				{
					$authResult['result']['user_id'] = $userId;
					$authResult['result']['client_id'] = $appId;

					static::store($authResult['result']);
				}

				return $authResult['result'];
			}
			else
			{
				return $authResult;
			}
		}

		return false;
	}

	public static function check($accessToken)
	{
		$managedCache = Application::getInstance()->getManagedCache();

		$authResult = false;
		if($managedCache->read(static::CACHE_TTL, static::getCacheId($accessToken)))
		{
			$authResult = $managedCache->get(static::getCacheId($accessToken));
		}

		if($authResult === false)
		{
			$client = OAuthService::getEngine()->getClient();
			$tokenInfo = $client->checkAuth($accessToken);

			if($tokenInfo['result'])
			{
				$authResult = $tokenInfo['result'];
				$authResult['user_id'] = $authResult['parameters'][static::PARAM_LOCAL_USER];
				unset($authResult['parameters'][static::PARAM_LOCAL_USER]);
			}
			else
			{
				$authResult = $tokenInfo;
				$authResult['access_token'] = $accessToken;
			}

			static::store($authResult);
		}
		elseif($authResult['expires_in'])
		{
			$authResult['expires_in'] -= time();
		}

		return $authResult;
	}

	public static function onRestCheckAuth(array $query, $scope, &$res)
	{
		$authKey = null;
		foreach(static::$authQueryParams as $key)
		{
			if(array_key_exists($key, $query))
			{
				$authKey = $query[$key];
				break;
			}
		}

		if($authKey)
		{
			$tokenInfo = static::check($authKey);
			if(is_array($tokenInfo))
			{
				$error = array_key_exists('error', $tokenInfo);

				if(!$error)
				{
					\CRestUtil::updateAppStatus($tokenInfo);
				}

				if(!$error && $tokenInfo['expires_in'] <= 0)
				{
					$tokenInfo = array('error' => 'expired_token');
					$error = true;
				}

				if(!$error && $scope !== \CRestUtil::GLOBAL_SCOPE && isset($tokenInfo['scope']))
				{
					$tokenScope = explode(',', $tokenInfo['scope']);
					if(!in_array($scope, $tokenScope))
					{
						$tokenInfo = array('error' => 'insufficient_scope');
						$error = true;
					}
				}

				if(!$error && $tokenInfo['user_id'] > 0)
				{
					if(!\CRestUtil::makeAuth($tokenInfo))
					{
						$tokenInfo = array('error' => 'authorization_error');
						$error = true;
					}
				}

				$res = $tokenInfo;

				$res['parameters_clear'] = static::$authQueryParams;

				return !$error;
			}

			return false;
		}

		return null;
	}

	protected static function store(array $authResult)
	{
		$managedCache = Application::getInstance()->getManagedCache();

		if($managedCache->read(static::CACHE_TTL, static::getCacheId($authResult["access_token"])))
		{
			$authResult['expires_in'] += time();
			$managedCache->set(static::getCacheId($authResult["access_token"]), $authResult);
		}
	}

	protected static function getCacheId($accessToken)
	{
		return static::CACHE_PREFIX.$accessToken;
	}

	protected static function getTokenParams($additionalParams, $userId)
	{
		if(!is_array($additionalParams))
		{
			$additionalParams = array();
		}

		$additionalParams[static::PARAM_LOCAL_USER] = $userId;
		$additionalParams[static::PARAM_TZ_OFFSET] = \CTimeZone::getOffset();

		return $additionalParams;
	}

}